/* SBUS Gamepad Converter
 * 
 * A sketch for the ESP32 to take SBUS input using RX2 and map it to a virtual bluetooth gamepad
 * 
 * Jump to line #87 to configure WiFi
 * 
 * Brian Dunigan - 2022
 * openstatic.org
 */
#include "sbus.h"
#include <Arduino.h>
#include <BleGamepad.h>
#include <WiFi.h>
#include <WiFiMulti.h>
#include <WebServer.h>
#include <WebSocketsServer.h>
#include <ArduinoJson.h>
#include <FFat.h>

BleGamepad bleGamepad("SBUS Gamepad", "openstatic", 100);
WiFiMulti wifiMulti;
WebSocketsServer webSocketServer(81);
WebServer httpServer(80);
uint8_t wl_status;

int numOfButtons = 8;
int numOfHatSwitches = 0;
boolean enableX = false;
boolean enableY = false;
boolean enableZ = false; // Cant exist with throttle
boolean enableRZ = false; // cant exist with rudder
boolean enableRX = false;
boolean enableRY = false;
boolean enableSlider1 = false;
boolean enableSlider2 = false;
boolean enableRudder = false;
boolean enableThrottle = false;
boolean enableAccelerator = false;
boolean enableBrake = false;
boolean enableSteering = false;
boolean enableStart = false;
boolean enableSelect = false;
String profile = "default";

long lastDecisecond = 0;

struct ChannelConfig
{
  String output;
  boolean inverted;
  int minimum;
  int maximum;
  int axis;
} CHANNELS[16];

const int buttonPin = 0;
int previousButtonState = HIGH;

/* SbusRx object on Serial1 */
bfs::SbusRx sbus_rx(&Serial2);
/* SbusTx object on Serial1 */
//bfs::SbusTx sbus_tx(&Serial2);
/* Array for storing SBUS data */
std::array<int16_t, bfs::SbusRx::NUM_CH()> sbus_data;

void handleNotFound();
void handleRoot();

void setup()
{
  pinMode(2, OUTPUT);
  for(int i = 0; i < 16; i++)
  {
    CHANNELS[i].output = "";
    CHANNELS[i].inverted = false;
    CHANNELS[i].minimum = 0;
    CHANNELS[i].maximum = 2000;
    CHANNELS[i].axis = 0;
    sbus_data[i] = 0;
  }
  /* Serial to display the received data */
  Serial.begin(115200);

  WiFi.mode(WIFI_STA);
  WiFi.setHostname("SBUSGAMEPAD");
  // put your setup code here, to run once:
  wifiMulti.addAP("YOUR NETWORK", "WIFI PASS");
  
  if (FFat.begin())
  {
    Serial.println("FFAT SUCCEED");
  } else {
    Serial.println("FFAT FAILED");
  }
  
  sbus_rx.Begin(16, 17);
  //sbus_tx.Begin(16, 17);

  loadProfile();

  Serial.println("CONFIGURING GAMEPAD");
  BleGamepadConfiguration bleGamepadConfig;
  bleGamepadConfig.setAutoReport(false);
  bleGamepadConfig.setControllerType(CONTROLLER_TYPE_GAMEPAD); // CONTROLLER_TYPE_JOYSTICK, CONTROLLER_TYPE_GAMEPAD (DEFAULT), CONTROLLER_TYPE_MULTI_AXIS
  bleGamepadConfig.setButtonCount(numOfButtons);
  bleGamepadConfig.setIncludeStart(enableStart);
  bleGamepadConfig.setIncludeSelect(enableSelect);
  bleGamepadConfig.setWhichAxes(enableX, enableY, enableZ, enableRX, enableRY, enableRZ, enableSlider1, enableSlider2);      // Can also be done per-axis individually. All are true by default
  bleGamepadConfig.setWhichSimulationControls(enableRudder, enableThrottle, enableAccelerator, enableBrake, enableSteering); // Can also be done per-control individually. All are false by default
  bleGamepadConfig.setHatSwitchCount(numOfHatSwitches);                                                                      // 1 by default
  bleGamepad.begin(&bleGamepadConfig);
  Serial.println("GAMEPAD OK");
  webSocketServer.begin();
  webSocketServer.onEvent(webSocketServerEvent);
  Serial.println("WSS Started");
  httpServer.on("/", handleRoot);
  httpServer.onNotFound(handleNotFound);
  httpServer.begin();
  Serial.println("HTTP Started");
  Serial.println("INIT COMPLETE");
  digitalWrite(2, HIGH);
}

void processJSONPayload(int num, uint8_t * payload)
{
  char* data = (char *) payload;
  DynamicJsonDocument root(4096);
  DeserializationError error = deserializeJson(root, data);
  if (error)
  {
    Serial.println(error.c_str());
  } else {
    if (root.containsKey("channels"))
    {
      JsonArray channels = root["channels"];
      for(int i = 0; i < channels.size(); i++)
      {
        JsonObject chan = channels[i];
        CHANNELS[i] = {
                        chan["output"],
                        chan["inv"],
                        chan["min"],
                        chan["max"]
                      };
      }
    }
    if (root.containsKey("profile"))
    {
      String newProfile = root["profile"].as<String>();
      if (FFat.exists("/" + profile + ".json") && !profile.equals(newProfile))
      {
        Serial.println("PROFILE CHANGED");
        profile = newProfile;
        if (!root.containsKey("save"))
          loadProfile();
      }
    }
    if (root.containsKey("save"))
    {
      saveProfile();
    }
    if (root.containsKey("apply"))
    {
      ESP.restart();
    }
  }
}

void doButton(int num, int adjustedAxis)
{
  if (adjustedAxis > 0)
  {
    bleGamepad.press(num);
  } else {
    bleGamepad.release(num);
  }
}

void loop() 
{
  wl_status = wifiMulti.run();
  if (wl_status == WL_CONNECTED)
  {
    webSocketServer.loop();
    httpServer.handleClient();
  }
  long ts = millis();
  if (ts - lastDecisecond >= 100)
  {
    if (wl_status == WL_CONNECTED)
    {
      everyNetworkDeciSecond();
    }
    lastDecisecond = ts;
  }
  if (sbus_rx.Read())
  {
    //Serial.println("SBUS READ");
    /* Grab the received data */
    sbus_data = sbus_rx.ch();
    boolean zeroPacket = true;
    boolean failsafePacket = sbus_rx.failsafe();
    boolean lostFrame = sbus_rx.lost_frame();
    for(int i = 0; i < 16; i++)
    {
      if (sbus_data[i] > 0)
        zeroPacket = false;
    }
    if (zeroPacket)
      Serial.println("ZERO PACKET");
    if (lostFrame)
      Serial.println("LOST FRAME");
    if (failsafePacket)
      Serial.println("FAILSAFE");
    /* Display the received data */
    if (!failsafePacket && !lostFrame && !zeroPacket)
    {
      for (int i = 0; i < 16; i++)
      {
        int adjustedAxis = CHANNELS[i].inverted ? map(sbus_data[i], CHANNELS[i].minimum, CHANNELS[i].maximum, 32767, -32767) : map(sbus_data[i], CHANNELS[i].minimum, CHANNELS[i].maximum, -32767, 32767);
        if (adjustedAxis < -32767)
          adjustedAxis = -32767;
        if (adjustedAxis > 32767)
          adjustedAxis = 32767;
        CHANNELS[i].axis = adjustedAxis;
        if (CHANNELS[i].output == "x")
        {
          bleGamepad.setX(adjustedAxis);
        } else if (CHANNELS[i].output == "y") {
          bleGamepad.setY(adjustedAxis);
        } else if (CHANNELS[i].output == "z") {
          bleGamepad.setZ(adjustedAxis);
        } else if (CHANNELS[i].output == "rx") {
          bleGamepad.setRX(adjustedAxis);
        } else if (CHANNELS[i].output == "ry") {
          bleGamepad.setRY(adjustedAxis);
        } else if (CHANNELS[i].output == "rz") {
          bleGamepad.setRZ(adjustedAxis);
        } else if (CHANNELS[i].output == "throttle") {
          bleGamepad.setThrottle(adjustedAxis);
        } else if (CHANNELS[i].output == "rudder") {
          bleGamepad.setRudder(adjustedAxis);
        } else if (CHANNELS[i].output == "slider1") {
          bleGamepad.setSlider1(adjustedAxis);
        } else if (CHANNELS[i].output == "slider2") {
          bleGamepad.setSlider2(adjustedAxis);
        } else if (CHANNELS[i].output == "brake") {
          bleGamepad.setBrake(adjustedAxis);
        } else if (CHANNELS[i].output == "steering") {
          bleGamepad.setSteering(adjustedAxis);
        } else if (CHANNELS[i].output == "accelerator") {
          bleGamepad.setAccelerator(adjustedAxis);
        } else if (CHANNELS[i].output == "button1") {
          doButton(1, adjustedAxis);
        } else if (CHANNELS[i].output == "button2") {
          doButton(2, adjustedAxis);
        } else if (CHANNELS[i].output == "button3") {
          doButton(3, adjustedAxis);
        } else if (CHANNELS[i].output == "button4") {
          doButton(4, adjustedAxis);
        } else if (CHANNELS[i].output == "button5") {
          doButton(5, adjustedAxis);
        } else if (CHANNELS[i].output == "button6") {
          doButton(6, adjustedAxis);
        } else if (CHANNELS[i].output == "button7") {
          doButton(7, adjustedAxis);
        } else if (CHANNELS[i].output == "button8") {
          doButton(8, adjustedAxis);
        } else if (CHANNELS[i].output == "start") {
          if (adjustedAxis > 0)
            bleGamepad.pressStart();
          else
            bleGamepad.releaseStart();
        } else if (CHANNELS[i].output == "select") {
          if (adjustedAxis > 0)
            bleGamepad.pressSelect();
          else
            bleGamepad.releaseSelect();
        }
      }
      if (bleGamepad.isConnected())
        bleGamepad.sendReport();
    }
  }
}

void shareProfiles(uint8_t num)
{
  String out;
  StaticJsonDocument<2048> jsonBuffer;
  JsonArray profiles = jsonBuffer.createNestedArray("profiles");
  File root = FFat.open("/");
  File file = root.openNextFile();
  while(file)
  {
    String filename = file.name();
    if (filename.endsWith(".json"))
    {
      profiles.add(filename.substring(0, filename.length() - 5));
    }
    file = root.openNextFile();
  }
  jsonBuffer["profile"] = profile;
  serializeJson(jsonBuffer, out);
  webSocketServer.sendTXT(num, out);
}

void shareChannels(uint8_t num)
{
  String out;
  StaticJsonDocument<2048> jsonBuffer;
  JsonArray controls = jsonBuffer.createNestedArray("channels");
  for(int i = 0; i < 16; i++)
  {
    JsonObject nested = controls.createNestedObject();
    nested["output"] = CHANNELS[i].output;
    nested["inv"] = CHANNELS[i].inverted;
    nested["min"] = CHANNELS[i].minimum;
    nested["max"] = CHANNELS[i].maximum;
    nested["val"] = sbus_data[i];
    nested["axis"] = CHANNELS[i].axis;
  }
  jsonBuffer["profile"] = profile;
  serializeJson(jsonBuffer, out);
  webSocketServer.sendTXT(num, out);
}

void everyNetworkDeciSecond()
{
  broadcastChannels();
}

void broadcastChannels()
{
  String out;
  StaticJsonDocument<2048> jsonBuffer;
  JsonArray controls = jsonBuffer.createNestedArray("channels");
  for(int i = 0; i < 16; i++)
  {
    JsonObject nested = controls.createNestedObject();
    nested["output"] = CHANNELS[i].output;
    nested["inv"] = CHANNELS[i].inverted;
    nested["min"] = CHANNELS[i].minimum;
    nested["max"] = CHANNELS[i].maximum;
    nested["val"] = sbus_data[i];
    nested["axis"] = CHANNELS[i].axis;
  }
  jsonBuffer["profile"] = profile;
  serializeJson(jsonBuffer, out);
  webSocketServer.broadcastTXT(out);
}

void webSocketServerEvent(uint8_t num, WStype_t type, uint8_t * payload, size_t lenght)
{
  switch (type)
  {
    case WStype_DISCONNECTED:
      break;
    case WStype_CONNECTED: {
        IPAddress ip = webSocketServer.remoteIP(num);
        Serial.print("Websocket Connection: ");
        Serial.println(ip);
        shareProfiles(num);
        shareChannels(num);
      }
      break;
    case WStype_TEXT:
      if (payload[0] == '{') {
        processJSONPayload(num, payload);
      }
      break;
    }
}


// ====================================================================================
// File Serving
// ====================================================================================


String getContentType(String filename)
{
  if (filename.endsWith(".htm")) {
    return "text/html";
  } else if (filename.endsWith(".html")) {
    return "text/html";
  } else if (filename.endsWith(".css")) {
    return "text/css";
  } else if (filename.endsWith(".js")) {
    return "application/javascript";
  } else if (filename.endsWith(".png")) {
    return "image/png";
  } else if (filename.endsWith(".gif")) {
    return "image/gif";
  } else if (filename.endsWith(".jpg")) {
    return "image/jpeg";
  } else if (filename.endsWith(".ico")) {
    return "image/x-icon";
  } else if (filename.endsWith(".svg")) {
    return "image/svg+xml";
  } else if (filename.endsWith(".xml")) {
    return "text/xml";
  } else if (filename.endsWith(".pdf")) {
    return "application/x-pdf";
  } else if (filename.endsWith(".zip")) {
    return "application/x-zip";
  } else if (filename.endsWith(".gz")) {
    return "application/x-gzip";
  }
  return "text/plain";
}

void handleFileRead(String path)
{
  bool found = false;
  if (path.endsWith("/"))
  {
    path += "index.html";
  }
  String contentType = getContentType(path);
  String pathWithGz = path + ".gz";
  if (FFat.exists(pathWithGz) || FFat.exists(path))
  {
    if (FFat.exists(pathWithGz))
    {
      path += ".gz";
    }
    File file = FFat.open(path, "r");
    httpServer.streamFile(file, contentType);
    file.close();
    found = true;
  }
  if (!found)
  {
    String message = "File Not Found\n\n";
    message += "URI: ";
    message += httpServer.uri();
    message += "\nMethod: ";
    message += ( httpServer.method() == HTTP_GET ) ? "GET" : "POST";
    message += "\nArguments: ";
    message += httpServer.args();
    message += "\n";
    message += "Path: ";
    message += path;
    message += "\n";
    
    for ( uint8_t i = 0; i < httpServer.args(); i++ )
    {     
      message += " " + httpServer.argName ( i ) + ": " + httpServer.arg ( i ) + "\n";
    }
    httpServer.send ( 404, "text/plain", message );
  }
}

// Load Settings from flash
void loadProfile()
{
  Serial.println("LOADING SETTINGS");
  StaticJsonDocument<2048> settings;
  File file = FFat.open("/" + profile + ".json", "r");
  if (!file)
  {
    Serial.println("/" + profile + ".json doesn't exist");
  } else {
    Serial.println("/" + profile + ".json found");
    DeserializationError error = deserializeJson(settings, file);
    if (error)
    {
      Serial.println("/" + profile + ".json corrupted"); 
    } else {
      Serial.println("/" + profile + ".json read ok!");
      file.close();
      if (settings.containsKey("channels"))
      {
        JsonArray channels = settings["channels"];
        for(int i = 0; i < channels.size(); i++)
        {
          JsonObject chan = channels[i];
          if (chan["output"] == "x")
          {
            enableX = true;
            Serial.println("enableX");
          } else if (chan["output"] == "y") {
            enableY = true;
            Serial.println("enableY");
          } else if (chan["output"] == "z") {
            enableZ = true;
            Serial.println("enableZ");
          } else if (chan["output"] == "rx") {
            enableRX = true;
            Serial.println("enableRX");
          } else if (chan["output"] == "ry") {
            enableRY = true;
            Serial.println("enableRY");
          } else if (chan["output"] == "rz") {
            enableRZ = true;
            Serial.println("enableRZ");
          } else if (chan["output"] == "rudder") {
            enableRudder = true;
            Serial.println("enableRudder");
          } else if (chan["output"] == "throttle") {
            enableThrottle = true;
            Serial.println("enableThrottle");
          } else if (chan["output"] == "slider1") {
            enableSlider1 = true;
            Serial.println("enableSlider1");
          } else if (chan["output"] == "slider2") {
            enableSlider2 = true;
            Serial.println("enableSlider2");
          } else if (chan["output"] == "steering") {
            enableSteering = true;
            Serial.println("enableSteering");
          } else if (chan["output"] == "brake") {
            enableBrake = true;
            Serial.println("enableBrake");
          } else if (chan["output"] == "accelerator") {
            enableAccelerator = true;
            Serial.println("enableAccelerator");
          } else if (chan["output"] == "start") {
            enableStart = true;
            Serial.println("enableStart");
          } else if (chan["output"] == "select") {
            enableSelect = true;
            Serial.println("enableSelect");
          }
          CHANNELS[i] = {
                          chan["output"],
                          chan["inv"],
                          chan["min"],
                          chan["max"]
                        };
        }
      }
    }
  }
}


void saveProfile()
{
  String out;
  StaticJsonDocument<2048> jsonBuffer;
  JsonArray controls = jsonBuffer.createNestedArray("channels");
  for(int i = 0; i < 16; i++)
  {
    JsonObject nested = controls.createNestedObject();
    nested["output"] = CHANNELS[i].output;
    nested["inv"] = CHANNELS[i].inverted;
    nested["min"] = CHANNELS[i].minimum;
    nested["max"] = CHANNELS[i].maximum;
  }
  Serial.println("SAVING PROFILE " + profile);
  File file = FFat.open("/" + profile + ".json", FILE_WRITE);
  if (!file) {
    Serial.println("COULDN'T WRITE FLASH");
    return;
  } else {
    serializeJson(jsonBuffer, file);
    file.flush();
    file.close();
    Serial.println("WROTE /" + profile + ".json");
  }
}

void handleRoot()
{
  handleFileRead(httpServer.uri());
}

void handleNotFound()
{
  handleFileRead(httpServer.uri());
}
